

from .hdl_tokenizer import _tokenize as hdl_tokenize
from .hdl_parser import HdlParser, Tokenizer
from ast import *

__all__ = ['hdl_tokenize', 'hdl_parse', 'hdl_compile', 'ast_dump']

def wrapped_tokenizer(source:str, verbose:bool = False) -> Tokenizer:

    # FIXME preprocess

    return Tokenizer(hdl_tokenize(source), verbose=verbose)


def hdl_parse(source:str, filename="<string>", py_version = None, verbose = False):
    tokenizer = wrapped_tokenizer(source)
    ast_tree = HdlParser(tokenizer, verbose=verbose, filename = filename, py_version=py_version).parse()
    return ast_tree

def hdl_compile(source:str, filename:str):
    
    ast_tree = hdl_parse(source, filename) 
    # FIXME postprocess
    code_obj = compile(ast_tree, filename, 'exec') 
    
    return code_obj






# see https://bitbucket.org/takluyver/greentreesnakes/src/master/astpp.py
def ast_dump(node, annotate_fields=True, include_attributes=False, indent='  '):
    """
    Return a formatted dump of the tree in *node*.  This is mainly useful for
    debugging purposes.  The returned string will show the names and the values
    for fields.  This makes the code impossible to evaluate, so if evaluation is
    wanted *annotate_fields* must be set to False.  Attributes such as line
    numbers and column offsets are not dumped by default.  If this is wanted,
    *include_attributes* can be set to True.
    """
    def _format(node, level=0):
        if isinstance(node, AST):
            fields = [(a, _format(b, level)) for a, b in iter_fields(node)]
            if include_attributes and node._attributes:
                fields.extend([(a, _format(getattr(node, a), level))
                               for a in node._attributes])
            return ''.join([
                node.__class__.__name__,
                '(',
                ', '.join(('%s=%s' % field for field in fields)
                           if annotate_fields else
                           (b for a, b in fields)),
                ')'])
        elif isinstance(node, list):
            lines = ['[']
            lines.extend((indent * (level + 2) + _format(x, level + 2) + ','
                         for x in node))
            if len(lines) > 1:
                lines.append(indent * (level + 1) + ']')
            else:
                lines[-1] += ']'
            return '\n'.join(lines)
        return repr(node)

    if not isinstance(node, AST):
        raise TypeError('expected AST, got %r' % node.__class__.__name__)
    return _format(node)



import sys


class When:
    def __init__(self, signal):
        self.caller = sys._getframe(1).f_locals
        if '__hdl_condition__' not in self.caller: 
            self.caller['__hdl_condition__'] = CondFrame()
        self._frame = self.caller['__hdl_condition__']
        self._cond = signal
    def __enter__(self):
        self._frame.new()
        self._frame.tail().when(self._cond)
        self._frame.update()
    def __exit__(self, exc_type, exc_value, traceback):
        self._frame.back()
        self._frame.update()
    
class Otherwise:
    def __init__(self):
        self._frame = sys._getframe(1).f_locals['__hdl_condition__']
    def __enter__(self):
        self._frame.forward()
        self._frame.tail().otherwise()
        self._frame.update()
    def __exit__(self, exc_type, exc_value, traceback):
        self._frame.back()
        self._frame.update()
    
    
class Elsewhen:
    def __init__(self, signal):
        self._frame = sys._getframe(1).f_locals['__hdl_condition__']
        self._cond = signal
    def __enter__(self):
        self._frame.forward()
        self._frame.tail().elsewhen(self._cond)
        self._frame.update()
    def __exit__(self, exc_type, exc_value, traceback):
        self._frame.back()   
        self._frame.update()
    
class CondList:
    def __init__(self):
        self.stack = []   # List[tuple[signal, bool]]
    def when(self, signal):
        self.stack.append((signal, True))
    def elsewhen(self, signal):
        self.otherwise()
        self.when(signal)
    def otherwise(self):
        last, _ = self.stack.pop()
        self.stack.append((last, False))
        
    
class CondFrame:
    def __init__(self):
        self.condlists: List[CondList] = []  # level0, level1, level2
        self.cache: CondList = None 
        
        self.conds = []
        
    def new(self):
        self.condlists.append(CondList())
    def back(self):
        self.cache = self.condlists.pop()
    def forward(self):
        if self.cache is not None:
            self.condlists.append(self.cache)
            self.cache = None 
        else:
            raise Exception("condition does not match")
    def tail(self)->CondList:
        return self.condlists[-1]
    def update(self):
        self.conds = []
        for l in self.condlists:
            self.conds += l.stack
    
    def __str__(self):
        return str(['{}{}'.format('' if x else '!',signal.__hdl_name__()) for signal, x in self.conds])
    

        
        
         



from typing import Any, Dict, List, Optional, Tuple, Union

import re 
import math

def width_infer(a: int, signed = False)->int:
    ''' 
    signed = False:
        a >= 0: uint 
        a < 0 : sint 
    signed = True:
        a : sint
    '''
    if a < 0:
        return math.ceil(math.log2(-a)) + 1   
    else:
        width_uint = 1 if a < 2 else math.floor(math.log2(a)) + 1
        if not signed:
            return width_uint
        else:
            return 1 if a == 0 else width_uint + 1
        return width_uint + int(signed)



def str2int(a:str, signed = False)->tuple:
    ''' 
    $1111_0000  signed 2's comp, store as is 
    +8'hf       signed, width = 8
    -8'hff      error, overflow
    '''
    a = a.lower()
    if a[0] == '$':
        value, width = str2uint(a[1:])   
        if value >> (width-1):   # sign ext
            value |=  -(1<<width)
        return (value, width, True)
    elif a[0] in ['+','-']:
        value, width = str2uint(a[1:])
        value = value if a[0] == '+' else -value 
        required_width = width_infer(value, signed = True)
        if "'" in a :    # width defined but overflow
            assert width >= required_width, "overflow"
        else:
            width = required_width
        return (value, width, True)
    elif signed:  
        return str2int('+' + a)
    else:       # unsigned
        value, width = str2uint(a)
        return (value, width, False)

def str2uint(a:str)->tuple:
    """ 
    input:
        1111_0000   unsigned bin 
        b1111_0000  unsigned bin 
        8'b11       unsigned bin
            
        h12ab       unsigned hex 
        8'hff       unsigned hex 
        
        d123        dec 
        8'd123      dec
        
    return:
        (value, width)
    """
    a_ = a
    a = a.lower()
    width = None
    if r := re.match(r"(\d+)'(.+)", a):
        width = int(r.groups()[0])
        a = r.groups()[1] 
    if a[0] == 'd':
        a = re.sub(r'[^0-9]', '',a) 
        value = int(a)
        width = width if width is not None else width_infer(value)
    elif a[0] == 'h':
        a = re.sub(r'[^0-9a-f]', '',a)
        width = width if width is not None else len(a)*4
        value = int(a, 16)
    else:
        a = re.sub(r'[^01]', '',a)
        width = width if width is not None else len(a)
        value = int(a, 2)
        
    required_width = width_infer(value)
    assert required_width <= width, f'overflow:{a_}'
    return value, width


def int2str(value: int, width: int) -> str:
    return ('{:0>'+str(width)+'b}').format(value & ( (1 << width) - 1))


from typing import Any, Dict, List, Optional, Tuple, Union
import numpy as np
import copy

from pyhgl.hdl._hdl import *




class Vec(Container):
    """
    n-D array of any signal, hardware, Vec, Bundle, or other python types
    
    init: 
        array like, len >= 1
        save copys 
    
    """
    
    def __init__(self, array):
        
        super().__init__()
        
        if isinstance(array, list):
            array = np.array(array, dtype=object)

        assert isinstance(array, np.ndarray), "Vec input should be array like"
        assert len(array) > 0, "empty Vec"
                
        self._shape: Tuple[int] = array.shape
        elements = []
        for i in array.flat: 
            elements.append(copy.copy(i))
        self._value = np.array(elements, dtype=object).reshape(self._shape)
        
        
    def _new(self, array: list):
        assert len(array) == len(self), 'length mismatch'
        return Vec(np.array(array, dtype=object).reshape(self._shape))
        
    def _flat(self):
        return self._value.flat
        
    @staticmethod
    def Full(shape, value):
        return Vec(np.full(shape, value, dtype = object))


    def _setname(self, name):
        for i, v in enumerate(self._value.flat):
            if isinstance(v, (Signal, Container)):
                v._setname(name + '_' + str(i))

    def __hdl_emit__(self):
        ret = f"{self.__hdl_name__()}{self._shape}" + "{\n"
        for i in self._value.flat:
            if isinstance(i, HDL):
                ret += i.__str__(1)
            else:
                ret += f"  {i}\n"
        return  ret + '}'
    
    def __len__(self):
        return len(self._value.flat)
       
    def __getitem__(self, key):
        single = True 
        if type(key) is tuple:
            for i in key:
                if type(i) is not int:
                    single = False 
        elif type(key) is not int:
            single = False
        return copy.copy(self._value[key]) if single else Vec(self._value[key])
    
    def __setitem__(self, key, value):
        self._value.__setitem__(key, value)
    
    def __copy__(self):
        return Vec(self._value)


class Bundle(Container):
    '''
    container of any signal, hardware, Vec, Bundle, or other python types
    init:
        named arguments 
        store copy
    getitem: 
        support int, slice, string
        return copy
    setitem:
        only string or int for existed key
    setattr:
        only for existed key
        
    '''
    
    def __init__(self, **args):
        
        super().__init__()
        self._value = {}
        self._keys: List[str] = []
        
        for k, v in args.items():
            # name start with _ will be ignored
            if k[0] == '_':
                continue       
            self._value[k] = copy.copy(v)
            self._keys.append(k)
        self.__dict__.update(self._value)

    def _new(self, array: list):
        assert len(array) == len(self), 'length mismatch'
        new_dict = dict(zip(self._keys, array))
        return Bundle(**new_dict)
        
    def _flat(self):
        return self._value.values()
        

    def _setname(self, name):
        for k, v in self._value.items():
            if isinstance(v, (Signal, Container)):
                v._setname(name + '_' + k)
    # return readable string
    def __hdl_emit__(self):
        ret = self.__hdl_name__() + "{\n"
        for k, v in self._value.items():
            if isinstance(v, HDL):
                ret = ret + v.__str__(1, prefix = f"{k}: ")  
            else:
                ret += "  " + f"{k}: {v}\n"
        return ret + "}"
    
    def __len__(self):
        return len(self._keys)
    
    def __getitem__(self, key):
        if isinstance(key, str):
            return copy.copy(self._value[key])
        if isinstance(key, int):
            return copy.copy(self._value[self._keys[key]])
        if isinstance(key, slice):
            # return a bundle 
            new_dict = {k:self._value[k] for k in self._keys[key]}
            return Bundle(**new_dict)
        
        raise Exception(f"key error: {key}")
    
    def __setitem__(self, key, value):
        if isinstance(key, int):
            key = self._keys[key]
        assert key in self._keys, f"unknown key {key}"
        # update both dict
        self._value[key] = value
        self.__dict__[key] = value
    
    def __setattr__(self, key, value):
        # unexisted key is not allowed
        if key[0] == '_':
            object.__setattr__(self, key, value)
        elif key in self._keys:
            self._value[key] = value
            self.__dict__[key] = value
        else:
            raise Exception(f"unknown key {key}")
    
    def __copy__(self):
        return Bundle(**self._value)
        

    
def Flatten(x: Container) -> list:
    if not isinstance(x, Container):
        return [x]
    ret = []
    for i in x._flat():
        ret += Flatten(i)
    return ret

def Zip(*args):
    return Map(*args, f = lambda *x: x)


# TODO 
def Filter(f, x: Container): ...
    





# see pegen/parser.py
# base class of pegen parser

import itertools
import os
import sys
import ast
from abc import abstractmethod
from typing import (
    Any, Callable, ClassVar, Dict, Iterator, List, Literal, Optional, Tuple, Type, TypeVar, Union, NoReturn, cast
)

from pyhgl.parser._tokenizer import Mark, Tokenizer, exact_token_types, tokenize

# Singleton ast nodes, created once for efficiency
Load = ast.Load()
Store = ast.Store()
Del = ast.Del()

Node = TypeVar("Node")
FC = TypeVar("FC", ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)

EXPR_NAME_MAPPING = {
    ast.Attribute: "attribute",
    ast.Subscript: "subscript",
    ast.Starred: "starred",
    ast.Name: "name",
    ast.List: "list",
    ast.Tuple: "tuple",
    ast.Lambda: "lambda",
    ast.Call: "function call",
    ast.BoolOp: "expression",
    ast.BinOp: "expression",
    ast.UnaryOp: "expression",
    ast.GeneratorExp: "generator expression",
    ast.Yield: "yield expression",
    ast.YieldFrom: "yield expression",
    ast.Await: "await expression",
    ast.ListComp: "list comprehension",
    ast.SetComp: "set comprehension",
    ast.DictComp: "dict comprehension",
    ast.Dict: "dict literal",
    ast.Set: "set display",
    ast.JoinedStr: "f-string expression",
    ast.FormattedValue: "f-string expression",
    ast.Compare: "comparison",
    ast.IfExp: "conditional expression",
    ast.NamedExpr: "named expression",
}



#################################################




T = TypeVar("T")
P = TypeVar("P", bound="Parser")
F = TypeVar("F", bound=Callable[..., Any])


def logger(method: F) -> F:
    """For non-memoized functions that we want to be logged.

    (In practice this is only non-leader left-recursive functions.)
    """
    method_name = method.__name__

    def logger_wrapper(self: P, *args: object) -> T:
        if not self._verbose:
            return method(self, *args)
        argsr = ",".join(repr(arg) for arg in args)
        fill = "  " * self._level
        print(f"{fill}{method_name}({argsr}) .... (looking at {self.showpeek()})")
        self._level += 1
        tree = method(self, *args)
        self._level -= 1
        print(f"{fill}... {method_name}({argsr}) --> {tree!s:.200}")
        return tree

    logger_wrapper.__wrapped__ = method  # type: ignore
    return cast(F, logger_wrapper)


def memoize(method: F) -> F:
    """Memoize a symbol method."""
    method_name = method.__name__

    def memoize_wrapper(self: P, *args: object) -> T:
        mark = self._mark()
        key = mark, method_name, args
        # Fast path: cache hit, and not verbose.
        if key in self._cache and not self._verbose:
            tree, endmark = self._cache[key]
            self._reset(endmark)
            return tree
        # Slow path: no cache hit, or verbose.
        verbose = self._verbose
        argsr = ",".join(repr(arg) for arg in args)
        fill = "  " * self._level
        if key not in self._cache:
            if verbose:
                print(f"{fill}{method_name}({argsr}) ... (looking at {self.showpeek()})")
            self._level += 1
            tree = method(self, *args)
            self._level -= 1
            if verbose:
                print(f"{fill}... {method_name}({argsr}) -> {tree!s:.200}")
            endmark = self._mark()
            self._cache[key] = tree, endmark
        else:
            tree, endmark = self._cache[key]
            if verbose:
                print(f"{fill}{method_name}({argsr}) -> {tree!s:.200}")
            self._reset(endmark)
        return tree

    memoize_wrapper.__wrapped__ = method  # type: ignore
    return cast(F, memoize_wrapper)


def memoize_left_rec(method: Callable[[P], Optional[T]]) -> Callable[[P], Optional[T]]:
    """Memoize a left-recursive symbol method."""
    method_name = method.__name__

    def memoize_left_rec_wrapper(self: P) -> Optional[T]:
        mark = self._mark()
        key = mark, method_name, ()
        # Fast path: cache hit, and not verbose.
        if key in self._cache and not self._verbose:
            tree, endmark = self._cache[key]
            self._reset(endmark)
            return tree
        # Slow path: no cache hit, or verbose.
        verbose = self._verbose
        fill = "  " * self._level
        if key not in self._cache:
            if verbose:
                print(f"{fill}{method_name} ... (looking at {self.showpeek()})")
            self._level += 1

            # For left-recursive rules we manipulate the cache and
            # loop until the rule shows no progress, then pick the
            # previous result.  For an explanation why this works, see
            # https://github.com/PhilippeSigaud/Pegged/wiki/Left-Recursion
            # (But we use the memoization cache instead of a static
            # variable; perhaps this is similar to a paper by Warth et al.
            # (http://web.cs.ucla.edu/~todd/research/pub.php?id=pepm08).

            # Prime the cache with a failure.
            self._cache[key] = None, mark
            lastresult, lastmark = None, mark
            depth = 0
            if verbose:
                print(f"{fill}Recursive {method_name} at {mark} depth {depth}")

            while True:
                self._reset(mark)
                self.in_recursive_rule += 1
                try:
                    result = method(self)
                finally:
                    self.in_recursive_rule -= 1
                endmark = self._mark()
                depth += 1
                if verbose:
                    print(
                        f"{fill}Recursive {method_name} at {mark} depth {depth}: {result!s:.200} to {endmark}"
                    )
                if not result:
                    if verbose:
                        print(f"{fill}Fail with {lastresult!s:.200} to {lastmark}")
                    break
                if endmark <= lastmark:
                    if verbose:
                        print(f"{fill}Bailing with {lastresult!s:.200} to {lastmark}")
                    break
                self._cache[key] = lastresult, lastmark = result, endmark

            self._reset(lastmark)
            tree = lastresult

            self._level -= 1
            if verbose:
                print(f"{fill}{method_name}() -> {tree!s:.200} [cached]")
            if tree:
                endmark = self._mark()
            else:
                endmark = mark
                self._reset(endmark)
            self._cache[key] = tree, endmark
        else:
            tree, endmark = self._cache[key]
            if verbose:
                print(f"{fill}{method_name}() -> {tree!s:.200} [fresh]")
            if tree:
                self._reset(endmark)
        return tree

    memoize_left_rec_wrapper.__wrapped__ = method  # type: ignore
    return memoize_left_rec_wrapper


class Parser:
    """Parsing base class."""

    KEYWORDS: ClassVar[Tuple[str, ...]]

    SOFT_KEYWORDS: ClassVar[Tuple[str, ...]]

    def __init__(self, tokenizer: Tokenizer, *, verbose: bool = False):
        self._tokenizer = tokenizer
        self._verbose = verbose
        self._level = 0
        self._cache: Dict[Tuple[Mark, str, Tuple[Any, ...]], Tuple[Any, Mark]] = {}
        # Integer tracking wether we are in a left recursive rule or not. Can be useful
        # for error reporting.
        self.in_recursive_rule = 0
        # Pass through common tokenizer methods.
        self._mark = self._tokenizer.mark
        self._reset = self._tokenizer.reset

    @abstractmethod
    def start(self) -> Any:
        pass

    def showpeek(self) -> str:
        tok = self._tokenizer.peek()
        return f"{tok.start[0]}.{tok.start[1]}: {tokenize.tok_name[tok.type]}:{tok.string!r}"

    @memoize
    def name(self) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.type == tokenize.NAME and tok.string not in self.KEYWORDS:
            return self._tokenizer.getnext()
        return None

    @memoize
    def number(self) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.type == tokenize.NUMBER:
            return self._tokenizer.getnext()
        return None

    @memoize
    def string(self) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.type == tokenize.STRING:
            return self._tokenizer.getnext()
        return None

    @memoize
    def op(self) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.type == tokenize.OP:
            return self._tokenizer.getnext()
        return None

    @memoize
    def type_comment(self) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.type == tokenize.TYPE_COMMENT:
            return self._tokenizer.getnext()
        return None

    @memoize
    def soft_keyword(self) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.type == tokenize.NAME and tok.string in self.SOFT_KEYWORDS:
            return self._tokenizer.getnext()
        return None

    @memoize
    def expect(self, type: str) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.string == type:
            return self._tokenizer.getnext()
        if type in exact_token_types:
            if tok.type == exact_token_types[type]:
                return self._tokenizer.getnext()
        if type in tokenize.__dict__:
            if tok.type == tokenize.__dict__[type]:
                return self._tokenizer.getnext()
        if tok.type == tokenize.OP and tok.string == type:
            return self._tokenizer.getnext()
        return None

    def expect_forced(self, res: Any, expectation: str) -> Optional[tokenize.TokenInfo]:
        if res is None:
            raise self.make_syntax_error(f"expected {expectation}")
        return res

    def positive_lookahead(self, func: Callable[..., T], *args: object) -> T:
        mark = self._mark()
        ok = func(*args)
        self._reset(mark)
        return ok

    def negative_lookahead(self, func: Callable[..., object], *args: object) -> bool:
        mark = self._mark()
        ok = func(*args)
        self._reset(mark)
        return not ok

    def make_syntax_error(self, message: str, filename: str = "<unknown>") -> SyntaxError:
        tok = self._tokenizer.diagnose()
        return SyntaxError(message, (filename, tok.start[0], 1 + tok.start[1], tok.line))







#################################################

    

class Parser(Parser):

    #: Name of the source file, used in error reports
    filename : str

    def __init__(self,
        tokenizer: Tokenizer, *,
        verbose: bool = False,
        filename: str = "<unknown>",
        py_version: Optional[tuple] = None,
    ) -> None:
        super().__init__(tokenizer, verbose=verbose)
        self.filename = filename
        self.py_version = min(py_version, sys.version_info) if py_version else sys.version_info
        self._exception = None

    def parse(self) -> Optional[ast.AST]:
        res = self.file()
        if res is None:
            raise SyntaxError("invalid syntax")
        return res

    def check_version(self, min_version: Tuple[int, ...], error_msg: str, node: Node) -> Node:
        """Check that the python version is high enough for a rule to apply.

        """
        if self.py_version >= min_version:
            return node
        else:
            raise SyntaxError(
                f"{error_msg} is only supported in Python {min_version} and above."
            )

    def raise_indentation_error(self, msg) -> None:
        """Raise an indentation error."""
        raise IndentationError(msg)

    def get_expr_name(self, node) -> str:
        """Get a descriptive name for an expression."""
        # See https://github.com/python/cpython/blob/master/Parser/pegen.c#L161
        assert node is not None
        node_t = type(node)
        if node_t is ast.Constant:
            v = node.value
            if v in (None, True, False, Ellipsis):
                return str(v)
            else:
                return "literal"

        try:
            return EXPR_NAME_MAPPING[node_t]
        except KeyError:
            raise ValueError(
                f"unexpected expression in assignment {type(node).__name__} "
                f"(line {node.lineno})."
            )

    def set_expr_context(self, node, context):
        """Set the context (Load, Store, Del) of an ast node."""
        node.ctx = context
        return node

    def ensure_real(self, number_str: str):
        number = ast.literal_eval(number_str)
        if number is not complex:
            self.store_syntax_error("real number required in complex literal")
        return number

    def ensure_imaginary(self, number_str: str):
        number = ast.literal_eval(number_str)
        if number is not complex:
            self.store_syntax_error("imaginary  number required in complex literal")
        return number

    def generate_ast_for_string(self, tokens):
        """Generate AST nodes for strings."""
        err_msg = ''
        line = 1
        col_offset = 0
        source = ''
        for t in tokens:
            n_line = t.start[0] - line
            if n_line:
                col_offset = 0
            source += """\n""" * n_line + ' ' * (t.start[1] - col_offset) + t.string
            line, col_offset = t.end
        if source[0] == ' ':
            source = '(' + source[1:]
        else:
            source = '(' + source
        source += ')'
        try:
            m = ast.parse(source)
        except SyntaxError as e:
            err_msg = e.args[0]
            # Identify the line at which the error occurred to get a more
            # accurate line number
            for t in tokens:
                try:
                    m = ast.parse(t.string)
                except SyntaxError:
                    break

        # Avoid getting a triple nesting in the error report that does not
        # bring anything relevant to the traceback.
        if err_msg:
            self.store_syntax_error_known_location(err_msg, t)
            raise self._exception

        return m.body[0].value

    def extract_import_level(self, tokens: List[tokenize.TokenInfo]) -> int:
        """Extract the relative import level from the tokens preceding the module name.

        '.' count for one and '...' for 3.

        """
        level = 0
        for t in tokens:
            if t.string == ".":
                level += 1
            else:
                level += 3
        return level

    def set_decorators(self,
        target: FC,
        decorators: list
    ) -> FC:
        """Set the decorators on a function or class definition."""
        target.decorator_list = decorators
        return target

    def get_comparison_ops(self, pairs):
        return [op for op, _ in pairs]

    def get_comparators(self, pairs):
        return [comp for _, comp in pairs]

    def set_arg_type_comment(self, arg, type_comment):
        if type_comment or sys.version_info < (3, 9):
            arg.type_comment = type_comment
        return arg

    def make_arguments(self,
        pos_only: Optional[List[Tuple[ast.arg, None]]],
        pos_only_with_default: List[Tuple[ast.arg, Any]],
        param_no_default: Optional[List[Tuple[ast.arg, None]]],
        param_default: Optional[List[Tuple[ast.arg, Any]]],
        after_star: Optional[Tuple[Optional[ast.arg], List[Tuple[ast.arg, Any]], Optional[ast.arg]]]
    ) -> ast.arguments:
        """Build a function definition arguments."""
        defaults = (
            [d for _, d in pos_only_with_default if d is not None]
            if pos_only_with_default else
            []
        )
        defaults += (
            [d for _, d in param_default if d is not None]
            if param_default else
            []
        )

        pos_only = pos_only or pos_only_with_default

        # Because we need to combine pos only with and without default even
        # the version with no default is a tuple
        pos_only = [p for p, _ in pos_only]
        params = (param_no_default or []) + ([p for p, _ in param_default] if param_default else [])

        # If after_star is None, make a default tuple
        after_star = after_star or (None, [], None)

        return ast.arguments(
            posonlyargs=pos_only,
            args=params,
            defaults=defaults,
            vararg=after_star[0],
            kwonlyargs=[p for p, _ in after_star[1]],
            kw_defaults=[d for _, d in after_star[1]],
            kwarg=after_star[2]
        )

    def _store_syntax_error(
        self,
        message: str,
        start: Optional[Tuple[int, int]] = None,
        end: Optional[Tuple[int, int]] = None
    ) -> None:
        line_from_token = start is None and end is None
        if start is None or end is None:
            tok = self._tokenizer.diagnose()
            start = start or tok.start
            end = end or tok.end

        if line_from_token:
            line = tok.line
        else:
            # End is used only to get the proper text
            line = "\\n".join(
                self._tokenizer.get_lines(list(range(start[0], end[0] + 1)))
            )

        self._exception = SyntaxError(
            message,
            (self.filename, start[0], start[1], line)
        )

    def store_syntax_error(self, message: str) -> None:
        self._store_syntax_error(message)

    def make_syntax_error(self, message: str) -> None:
        self._store_syntax_error(message)
        return self._exception

    def store_syntax_error_known_location(self, message: str, node) -> None:
        """Store a syntax error that occured at a given AST node."""
        if isinstance(node, tokenize.TokenInfo):
            start = node.start
            end = node.end
        else:
            start = node.lineno, node.col_offset
            end = node.end_lineno, node.end_col_offset

        self._store_syntax_error(message, start, end)

    def store_syntax_error_known_range(
        self,
        message: str,
        start_node: Union[ast.AST, tokenize.TokenInfo],
        end_node: Union[ast.AST, tokenize.TokenInfo]
    ) -> None:
        if isinstance(start_node, tokenize.TokenInfo):
            start = start_node.start
        else:
            start = start_node.lineno, start_node.col_offset

        if isinstance(end_node, tokenize.TokenInfo):
            end = end_node.end
        else:
            end = end_node.end_lineno, end_node.end_col_offset

        self._store_syntax_error(message, start, end)

    def store_syntax_error_starting_from(
        self,
        message: str,
        start_node: Union[ast.AST, tokenize.TokenInfo]
    ) -> None:
        if isinstance(start_node, tokenize.TokenInfo):
            start = start_node.start
        else:
            start = start_node.lineno, start_node.col_offset

        self._store_syntax_error(message, start, None)

    def raise_syntax_error(self, message: str) -> NoReturn:
        self._store_syntax_error(message)
        raise self._exception

    def raise_syntax_error_known_location(
            self,
            message: str,
            node: Union[ast.AST, tokenize.TokenInfo]
        ) -> NoReturn:
        """Raise a syntax error that occured at a given AST node."""
        self.store_syntax_error_known_location(message, node)
        raise self._exception

    def raise_syntax_error_known_range(
        self,
        message: str,
        start_node: Union[ast.AST, tokenize.TokenInfo],
        end_node: Union[ast.AST, tokenize.TokenInfo]
    ) -> NoReturn:
        self.store_syntax_error_known_range(message, start_node, end_node)
        raise self._exception

    def raise_syntax_error_starting_from(
        self,
        message: str,
        start_node: Union[ast.AST, tokenize.TokenInfo]
    ) -> NoReturn:
        self.store_syntax_error_starting_from(message, start_node)
        raise self._exception


from .Logic import *


class SInt(Logic):
    ''' 
    '''
    def __init__(self,value: Union[int, str, bool] = 0, w: int = None):
        super().__init__()
        # basic attribute
        self.value: int = 0 
        self.width: int = 1 
        self.shape: Tuple[int, ...] = None
        # leaf node ? 
        self.driver = None   
        self._name = ""
        # for input output 
        self._flip = False 
        self._direction = 'inner'
        
        if isinstance(value, (bool,int)):
            self.value = int(value)
            self.width = utils.width_infer(self.value, signed=True)
        elif isinstance(value, str):
            self.value, self.width, _ = utils.str2int(value, signed=True)
        else:
            raise Exception("invalid signal input")

        if w is not None:
            assert w >= self.width, f'overflow: {value} with width {w}'
            self.width = w
    
        
    # TODO
    # def __getitem__(self, key):
    #     if isinstance(key, str):
    #         return copy.copy(self.value[key])
    #     if isinstance(key, int):
    #         return copy.copy(self.value[self._keys[key]])
    #     if isinstance(key, slice):
    #         # return a bundle 
    #         new_dict = {}
    #         for k in self._keys[key]:
    #             new_dict[k] = self.value[k]
    #         return Bundle(**new_dict)
    #     raise Exception(f"key error: {key}")

    
    # def __deepcopy__(self,memo):
    #     cls = self.__class__
    #     ret = cls.__new__(cls)
    #     ret.__dict__.update(self.__dict__)
    #     ret.driver = None 
    #     ret._name = ""
    #     return ret
 


from typing import Any, Dict, List, Optional, Tuple, Union

from pyhgl.hdl._hdl import * 
from pyhgl.hdl import utils

class Logic(Signal):
    '''
    1-D bits
    -------- 
     
    Attributes
    - value, width: basic
    - shape: slice
    - driver = None: leafNode
    - _name: basic
    
    
    '''
    def __init__(self,value: Union[int, str, bool] = 0, w: int = None ):
        super().__init__()
        
        if isinstance(value, (bool,int)):
            self.value = int(value)
            self.width = utils.width_infer(self.value)
        elif isinstance(value, str):
            self.value, self.width, _ = utils.str2int(value)
        else:
            raise Exception("invalid signal input")
        if w is not None:
            assert w >= self.width, f'overflow: {value} with width {w}'
            self.width = w
    
    def __hdl_name__(self):
        if self.driver is None:
            return f"{self.width}'b{utils.int2str(self.value, self.width)}" 
        else:
            return super().__hdl_name__()  
    
    def __hdl_emit__(self):
        value = f"{self.width}'b{utils.int2str(self.value, self.width)}" 
        if self.driver is None:
            return value
        else:
            return f"{self._direction} {self.__hdl_name__()} = {value}"

    def W(length: int):
        assert self.width > length 
        return Slice(self, length-1, 0) 
    
    # TODO
    def View(*shape): ...

    def __invert__(self): 
        return NOT(self)
    
    
    

    
class AND(Node):
    def __init__(self, a: Signal, b: Signal):
        super().__init__()
        if type(b) is str: b = Logic(b)
        self.inputs = [a, b]
        longer = a if a.width > b.width else b
        self.outputs = [longer._new(value=0)]
        self.outputs[0].driver = self

    def forward(self):
        self.outputs[0].value = self.inputs[0].value & self.inputs[1].value


class OR(Node):

    def __init__(self, a: Signal, b: Signal):
        super().__init__()
        if type(b) is str: b = Logic(b)
        self.inputs = [a, b]
        longer = a if a.width > b.width else b
        self.outputs = [longer._new(value=0)]
        self.outputs[0].driver = self

    def forward(self):
        self.outputs[0].value = self.inputs[0].value | self.inputs[1].value


class NOT(Node):

    def __init__(self, a: Signal):
        super().__init__()
        self.inputs = [a]
        self.outputs = [a._new(value=0)]
        self.outputs[0].driver = self

    def forward(self):
        self.outputs[0].value = ~self.inputs[0].value


class XOR(Node):

    def __init__(self, a: Signal, b: Signal):
        super().__init__()
        if type(b) is str: b = Logic(b)
        self.inputs = [a, b]
        longer = a if a.width > b.width else b
        self.outputs = [longer._new(value=0)]
        self.outputs[0].driver = self

    def forward(self):
        self.outputs[0].value = self.inputs[0].value ^ self.inputs[1].value




class Mux(Node):
    '''
    2 to 1 mux 
    sel != 0: first
    else: second
    '''

    def __init__(self, sel: Signal, first: Signal, second: Signal):
        super().__init__()
        if type(first) is str: first = Logic(first)
        if type(second) is str: second = Logic(second)
        self.inputs = [sel, first, second]
        longer = first if first.width > second.width else second
        self.outputs = [longer._new(value=0)]
        self.outputs[0].driver = self

    def forward(self):
        self.outputs[0].value = self.inputs[1].value if self.inputs[0].value != 0 else self.inputs[2].value


class Cat(Node):
    def __init__(self, *args: Signal):
        super().__init__()
        assert len(args) > 0, 'no input'
        self.inputs = []
        width = 0 
        for i in args:
            if type(i) is str: i = Logic(i)
            self.inputs.append(i)
            width += i.width
        self.outputs = [args[0]._new(value=0, width=width)]
        self.outputs[0].driver = self

    def forward(self):
        value = 0
        for i in self.inputs:
            value = (i.value & ((1 << i.width) - 1)) | (value << i.width)
        self.outputs[0].value = value


class Slice(Node):
    ''' 
    apply slice on 1-D bits 
    1111_0000
    ^^^^ ^^^^
    7654 3210
    
    return bits from start to end, including endpoint
    if start < end, reversed order
    '''
    def __init__(self, driver: Signal, start: int, end: int = None):
        super().__init__()
        if not end:
            end = start
        assert start >= 0 and start < driver.width 
        assert end >= 0 and end < driver.width 
        
        width = abs(start - end) + 1
        self.inputs = [driver]
        self.outputs = [driver._new(value=0, width=width)]     
        self.outputs[0].driver = self
        self._start = start
        self._end = end

    def __hdl_emit__(self):
        left = self.outputs[0].__hdl_name__()
        right = self.inputs[0].__hdl_name__()
        return f"{left} = {right}[{self._start}:{self._end}]"


    def forward(self):
        value = self.inputs[0].value
        width = self.outputs[0].width
        n_shift = min(self._end, self._start)
        value = (value >> self._end) & ((1 << width) - 1)
        if self._start >= self._end:
            self.outputs[0].value = value 
        else:
            # reverse 
            real_value = 0 
            for i in range(width):
                real_value <<= 1
                real_value += value & 1 
                value >>= 1
            self.outputs[0].value = real_value 








from typing import Any, Dict, List, Optional, Tuple, Union

from pyhgl.hdl._hdl import *
from pyhgl.hdl import _global
from pyhgl.hdl.signal import Logic


"""   
basic hardware nodes
    Wire: signal pass 
    Reg: signal store 
    WireInOut
    Mux: 2 to 1 mux
"""


class Wire(Node):

    def __init__(self, driver: Signal):
        super().__init__()
        assert isinstance(driver, Signal)
        self.inputs = [driver]
        self.outputs = [driver._new(value=0)]    # no value, only width
        self.outputs[0].driver = self

    def __new__(cls, driver):
        if isinstance(driver, Container):
            return Map(driver, f = Wire)
        else:
            ret = object.__new__(cls)
            ret.__init__(driver)
            return ret.outputs[0]    # return Signal

    def forward(self):
        self.outputs[0].value = self.inputs[0].value


class Reg(Node):

    def __init__(self, driver: Signal, init: Signal = None, reset=None, clock=None, posedge=True):
        ''' 
        reset & clock: Signal
        '''
        super().__init__()
        assert isinstance(driver, Signal)
        if reset is None:
            reset: Signal = _global._reset
        if clock is None:
            clock: Signal = _global._clocks.clock
        if init is None:
            init = driver._new(value=0)

        self.inputs = [driver, init, reset, clock]
        self.outputs = [driver._new(value=0)]
        self.outputs[0].driver = self
        self.outputs[0].value = init.value
        # for step
        self.curr_value = None
        self.curr_clock = None
        self.prev_clock = None
        self.posedge = posedge

    def __new__(cls, driver, *args, **kwargs):
        if isinstance(driver, Container):
            return Map(driver, f = Reg)
        else:
            reg = object.__new__(cls)
            reg.__init__(driver, *args, **kwargs)
            return reg.outputs[0]    # return Signal

    def __hdl_emit__(self):
        left = self.outputs[0].__hdl_name__()
        driver, init, reset, clock = self.inputs
        return f"{left} = Reg({init.__hdl_name__()}, {reset.__hdl_name__()}, {clock.__hdl_name__()})"
    
    def __hdl_emit2__(self):
        return self.outputs[0].__hdl_name__() + ' <= ' + self.inputs[0].__hdl_name__()

    def forward(self): 
        self.curr_value = self.inputs[0].value
        self.prev_clock = self.curr_clock 
        self.curr_clock = self.inputs[3].value != 0
         
    def step(self): 
        if self.inputs[2].value != 0:  # reset
            self.outputs[0].value = self.outputs[1].value
            return
        if self.prev_clock is not None and self.curr_clock is not None:
            if self.posedge:
                if not self.prev_clock and self.curr_clock:
                    self.outputs[0].value = self.curr_value
            else:
                if self.prev_clock and not self.curr_clock:
                    self.outputs[0].value = self.curr_value
         

class WireInOut(Node):
    ''' 
    inputs: [sel, signal, sel, signal, ...]
    outputs: [signal, signal, ...]
    '''
    def __init__(self, x: Signal):
        super().__init__()
        assert x._direction == 'inout', 'inout wire needs inout signal'
        self.inputs = []
        self.outputs = [x._new(value=0)]
        self.outputs[0].driver = self
        self.default_value = x.value
        
    def __new__(cls, x):
        if isinstance(x, Container):
            return Map(x, f = WireInOut)
        else:
            ret = object.__new__(cls)
            ret.__init__(x)
            return ret.outputs[0]     
        
    def _merge(self, other):
        assert isinstance(other, WireInOut) and self.outputs[0].width == other.outputs[0].width 
        self.inputs += other.inputs
        self.outputs += other.outputs 
        for signal in self.inputs + self.outputs:
            signal.driver = self
        
    def forward(self):
        value = None 
        for i in range(len(self.inputs)/2):
            if self.inputs[i*2].value != 0:   # select that value
                if value is None:
                    value = self.inputs[i*2+1].value
                else:
                    raise Exception("simulation error: multiple driver of inout wire")
        if value is None:
            value = x.vlaue
        for i in self.outputs:
            i.value = value
        
        
        
def Tri(sel: Signal, driver: Signal, inout: Signal) -> Union[Signal, Container]:
    ''' 
    sel:   if sel, inout <= driver; else inout <= high-z

    return: output signal of the inout wire
    '''            
    if isinstance(inout, Container):
        return Map(sel, driver, inout, f=Tri)
    else:
        assert isinstance(inout, Signal) and inout._direction == 'inout' and isinstance(inout.driver, WireInOut)
        assert isinstance(sel, Signal)
        assert isinstance(driver, Signal)
        inout.driver.inputs.extend([sel, dirver])
        return inout.driver.outputs[0]
        
    



class _Mux(Node):
    '''
    2 to 1 mux used in conditional assignment
    do not use it explictly!
    if sel, return first
    '''

    def __init__(self, sel: Signal, first: Signal, second: Signal):
        super().__init__()
        self.inputs = [sel, first, second]
        longer = first if first.width > second.width else second
        self.outputs = [longer._new(value=0)]
        self.outputs[0].driver = self

    def forward(self):
        self.outputs[0].value = self.inputs[1].value if self.inputs[0].value != 0 else self.inputs[2].value


# conditional assignment

def _condwalker(inports: List[Signal], idx: int, conds:list, new_signal: Signal):
    ''' 
    inports[idx]: Wire.inputs[0] | Reg.inputs[0] | _Mux.inputs[1] | _Mux.inputs[2]
    
    conds: [(c0, True), (c1, False), ...]
    '''
    if len(conds) == 0:
        inports[idx] = new_signal 
        return 
    
    default_signal = inports[idx]
    cond_signal, cond_sel = conds[0]
    
    if isinstance(default_signal.driver, _Mux) and default_signal.driver.inputs[0] is cond_signal:
        pass # do not generate new mux 
    else:
        inports[idx] = _Mux(cond_signal, default_signal, default_signal)  # new mux takes default signal
    # goto _Mux
    _condwalker(inports[idx].driver.inputs, int(not cond_sel)+1, conds[1:], new_signal)



def _update(left: Union[Signal, Container], right: Union[Signal, Container, str], conds:list) -> None:
    ''' 
    left: Signal  or their Container  
    right: Signal or their Container  
    
    !!! notice: assignment operator should always return left side 
    '''
    if isinstance(left, Container):
        Map(left, right, conds, f = _update_element)
    # is logic
    _update_element(left, right, conds)
        
def _update_element(left: Signal, right: Union[Signal, str], conds: list) -> None:
    if type(right) is str: 
        right = Logic(right)
    assert isinstance(left, Signal) and isinstance(right, Signal), "update nonsignal"
    # conditional update
    if left._direction not in ['input', 'inout'] and right._direction != 'inout':
        _condwalker(left.driver.inputs, 0, conds, right) 
    # inout connection
    elif left._direction == 'inout' and right._direction == 'inout':
        left.driver._merge(right.driver) 
    else:
        raise Exception(f"Error: {left._direction} <<= {right._direction}")
    
__hdl_op__['__ilshift__'].default = _update





from pyhgl.hdl._hdl import *
from .Logic import *
from .SInt import *
from .UInt import *


__hdl_op__['__and__'].default = AND
__hdl_op__['__or__'].default = OR
__hdl_op__['__xor__'].default = XOR


# __hdl_op__['__and__'].register(Logic, AND, Logic, SInt, UInt, order='both')
# __hdl_op__['__and__'].register(SInt, AND, Logic, SInt, UInt, order='both')
# __hdl_op__['__and__'].register(UInt, AND, Logic, SInt, UInt, order='both')

# __hdl_op__['__or__'].register(Logic, OR, Logic, SInt, UInt, order='both')
# __hdl_op__['__or__'].register(SInt, OR, Logic, SInt, UInt, order='both')
# __hdl_op__['__or__'].register(UInt, OR, Logic, SInt, UInt, order='both')

# __hdl_op__['__xor__'].register(Logic, XOR, Logic, SInt, UInt, order='both')
# __hdl_op__['__xor__'].register(SInt, XOR, Logic, SInt, UInt, order='both')
# __hdl_op__['__xor__'].register(UInt, XOR, Logic, SInt, UInt, order='both')




""" 
python import hook: https://www.python.org/dev/peps/pep-0302/
reference: https://github.com/pfalcon/python-imphook
"""

import importlib
import sys 
import traceback

from pyhgl.parser import hdl_compile
from pyhgl import config



_hdlext = '.pyh'


class HdlFileLoader(importlib._bootstrap_external.FileLoader):

    def create_module(self, spec):
        # print(spec.name, spec.origin)
        # print(self.name, self.path)

        with open(self.path) as f:
            source_code = f.read()
        self.code_obj = hdl_compile(source_code, self.path)
        
        return None
        
    def exec_module(self, module):
        try:
            exec(self.code_obj, module.__dict__)  
        except:
            print(traceback.format_exc())

def add_import_hook():

    for i, path_hook in enumerate(sys.path_hooks):
        if not isinstance(path_hook, type):
            # Assume it's a type wrapped in a closure,
            # as is the case for FileFinder.
            path_hook = type(path_hook("."))
        if path_hook is importlib._bootstrap_external.FileFinder:
            sys.path_hooks.pop(i)
            insert_pos = i
            break
    else:
        config.logger.warning("Could not find existing FileFinder to replace, installing ours as the first to use")
        insert_pos = 0

    # Mirrors what's done by importlib._bootstrap_external._install(importlib._bootstrap)
    # loaders for different file extensions
    loaders = [(HdlFileLoader, [_hdlext])] + importlib._bootstrap_external._get_supported_file_loaders()
    # path_hook closure captures supported_loaders in itself, return FileFinder
    sys.path_hooks.insert(insert_pos, importlib._bootstrap_external.FileFinder.path_hook(*loaders))
    sys.path_importer_cache.clear()

add_import_hook() 






# -*- coding: utf-8; mode: python -*-
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyhgl",
    version="0.0.3",
    author="jintao",
    author_email="jintaos2@zju.edu.cn",
    description="a python like hdl",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=None,
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)



from typing import Any, Dict, List, Optional, Set, Tuple, Union
import sys

__all__ = ['HDL', 'Container', 'Map', '__hdl_op__', 'Signal', 'Node']

class HDL:
    ''' 
    internal methods
    ----------------
    '''
    def __init__(self): 
        self._name: str = ''
        
    def __hdl_version__(self): 
        return (0,0,1)
    
    def _setname(self, name: str = ''):         raise Exception("not implemented") 
    def __hdl_name__(self):                     raise Exception("not implemented") 
    
    # return string with intend and prefix 
    def __str__(self, intend = 0, prefix = ''):
        s = self.__hdl_emit__().splitlines(keepends=True)
        return "  "*intend + prefix + ("  "*intend).join(s) + '\n'
    def __hdl_emit__(self):                     raise Exception("not implemented")
    
    # because __bool__ can only return True/False, to return HDL, use: when (a == 0 | b != 0): ...
    def __bool__(self):                         raise Exception("bool is not allowed")     
    def __copy__(self): return self      
    def __hash__(self): return id(self)
    
    '''
    special operators & unary operators
    -----------------------------------
    '''
    def __getitem__(self, args):                raise Exception("not implemented")
    def __setitem__(self, key, value):          raise Exception("not implemented")
     # a**2   duplicate bits
    def __pow__(self, other):                   raise Exception("not implemented")       
    def __bytes__(self):                        raise Exception("not implemented")   
    
    def __len__(self):                          raise Exception("not implemented")           
    def __invert__(self):                       raise Exception("not implemented")           
    def __abs__(self):                          raise Exception("not implemented")           
    def __pos__(self):                          raise Exception("not implemented")           
    def __neg__(self):                          raise Exception("not implemented")               
    

    ''' 
    external binary operators
    -------------------------
    overload of python binary operators like: +-*/|&
    only for Logic/Contianer Type
    '''
    # bitwise
    def __or__(self, other):        return __hdl_op__['__or__'].apply(self, other)      # a | b
    def __and__(self, other):       return __hdl_op__['__and__'].apply(self, other)     # a & b
    def __xor__(self, other):       return __hdl_op__['__xor__'].apply(self, other)     # a ^ b
    def __lshift__(self, other):    return __hdl_op__['__lshift__'].apply(self, other)  # a << b
    def __rshift__(self, other):    return __hdl_op__['__rshift__'].apply(self, other)  # a >> b
    
    # used for comparison, return 1-bit signal
    def __lt__(self, other):        return __hdl_op__['__lt__'].apply(self, other)      # a < b
    def __gt__(self, other):        return __hdl_op__['__gt__'].apply(self, other)      # a > b
    def __le__(self, other):        return __hdl_op__['__le__'].apply(self, other)      # a <= b
    def __ge__(self, other):        return __hdl_op__['__ge__'].apply(self, other)      # a >= b
    def __eq__(self, other):        return __hdl_op__['__eq__'].apply(self, other)      # a == b
    def __ne__(self, other):        return __hdl_op__['__ne__'].apply(self, other)      # a != b
    # calculate
    def __add__(self, other):       return __hdl_op__['__add__'].apply(self, other)     # a + b
    def __sub__(self, other):       return __hdl_op__['__sub__'].apply(self, other)     # a - b
    def __mul__(self, other):       return __hdl_op__['__mul__'].apply(self, other)     # a * b
    def __matmul__(self, other):    return __hdl_op__['__matmul__'].apply(self, other)  # a @ b
    def __truediv__(self, other):   return __hdl_op__['__truediv__'].apply(self, other) # a / b
    def __mod__(self, other):       return __hdl_op__['__mod__'].apply(self, other)     # a % b
    def __floordiv__(self, other):  return __hdl_op__['__floordiv__'].apply(self,other) # a // b
    
    '''
    assignment & connection
    -----------------------
    always return self !!!
    '''
    
    # hdl assignment <<=, like assign a = b in systemverilog
    def __ilshift__(self, other):  
        caller = sys._getframe(1).f_locals
        if '__hdl_condition__' not in caller:
            conds = []  
        else:
            conds = caller['__hdl_condition__'].conds
        __hdl_op__['__ilshift__'].apply(self, other, conds)         
        return self  
    
    def __iadd__(self, other):      __hdl_op__['__iadd__'].apply(self, other);      return self # +=
    def __isub__(self, other):      __hdl_op__['__isub__'].apply(self, other);      return self # -=
    def __imul__(self, other):      __hdl_op__['__imul__'].apply(self, other);      return self # *=  
    def __imatmul__(self, other):   __hdl_op__['__imatmul__'].apply(self, other);   return self # @=           
    def __ifloordiv__(self, other): __hdl_op__['__ifloordiv__'].apply(self, other); return self # //=
    def __itruediv__(self, other):  __hdl_op__['__itruediv__'].apply(self, other);  return self # /=
    def __imod__(self, other):      __hdl_op__['__imod__'].apply(self, other);      return self # %= 
    def __ipow__(self,other):       __hdl_op__['__ipow__'].apply(self, other);      return self # **=
    def __irshift__(self, other):   __hdl_op__['__irshift__'].apply(self, other);   return self # >>=
    def __iand__(self, other):      __hdl_op__['__iand__'].apply(self, other);      return self # &=
    def __ior__(self, other):       __hdl_op__['__ior__'].apply(self, other);       return self # |=
    def __ixor__(self, other):      __hdl_op__['__ixor__'].apply(self, other);      return self # ^=



class Container(HDL):
    ''' 
    features:
      - len(a) return number of elements 
      - support index a[idx]
    '''
    def __init__(self):
        super().__init__()
        
    def __hdl_name__(self): return self.__class__.__name__  
    
    def _flat(self): ...  # return an itelator
    
    def _new(self, array: list): ... # input a list, return new instance with same shape
        
def Map(*args, f = None):
    ''' 
    map functions on containers 
    '''
    first = None
    for arg in args:
        if isinstance(arg, Container):
            if first is None:
                first = arg 
            else:
                assert len(first) == len(arg), "shape mismatch!"
    if first is None:   # no containers
        return f(*args)
    # extract containers
    new_elements = []
    for n in range(len(first)):
        arg_list = []
        for arg in args:
            if isinstance(arg, Container):
                arg_list.append(arg[n])
            else:
                arg_list.append(arg)
        new_elements.append(f(*arg_list))
    return first._new(new_elements)
        


    
'''
operator behavior:
unary_op:
    if not implemented, call base class 
binary_op:
    1. search registered functions for that operator
    2. apply default function if defined
    3. otherwise deal with container
    4. raise error
'''
class __hdl_op__:
    class Operator:
        def __init__(self, name:str):
            self.name = name 
            self.default = None
            self.table = {}    # a table of functions

        def find(self, l, r):
            if l not in self.table: self.table[l] = {}
            if r not in self.table[l]: self.table[l][r] = None     
            return self.table[l][r]    
           
        def _register(self, f, left: type, right: type):
            l = left.__name__
            r = right.__name__
            self.find(l, r)
            self.table[l][r] = f 
            
        def register(self, left: type, f, *right: type, order = 'normal' ):
            assert order in ['normal', 'inverse', 'both']
            if order in ['normal','both']:
                for r in right:
                    self._register(f, left, r)
            if order in ['inverse', 'both']:
                for r in right:
                    self._register(f, r, left)
            
        def apply(self, left, right, *args) -> Any:
            l = left.__class__.__name__
            r = right.__class__.__name__
            f = self.find(l, r)
            if f is not None: 
                return f(left, right, *args)
            if self.default is not None:
                return self.default(left, right, *args)
            if isinstance(left, Container) or isinstance(right, Container): 
                return Map(left, right, *args, f = lambda x,y: getattr(x, self.name)(y))
            raise Exception(f"{l}{self.name}{r} not defined")
        
        def __str__(self):
            return self.name + ' ' + str(self.table) + '\n'
            
    def __init__(self):
        self.op: Dict[str, __hdl_op__.Operator] = {}     
    def __getitem__(self, key:str):
        assert len(key) > 4 and key[:2] == '__' and key[-2:] == '__'
        if key not in self.op:
            self.op[key] = __hdl_op__.Operator(key)
        return self.op[key]
    def __str__(self):
        ret = ''
        for i in self.op.values():
            ret += str(i)
        return ret

__hdl_op__ = __hdl_op__()

__names__: Set[str] = set()

class Signal(HDL):
    def __init__(self, value: int = 0, width:int = 1):
        super().__init__()
        # basic attribute
        self.value: int = value 
        self.width: int = width
        self.driver: Node = None   
        # for io and assignment
        self._flip: bool = False 
        self._direction = 'inner'      
          
    def _set(self, value: int = 0):
        self.value = value       
         
    def _new(self, value: int = None, width: int = None):
        ''' 
        parameter: 
          1. value
          2. width
        return a new instance without other attributes
        like copy
        '''
        if value is None: value = self.value
        if width is None: width = self.width 
        return self.__class__(value, width)

    def _setname(self, name: str = ''):
        # do not set
        if len(self._name) > 0: 
            return self 
        # set name
        if len(name) > 0 and name not in __names__:
            self._name = name         
        else:
            if len(name) == 0:
                name = self.__class__.__name__            
            n = 1 
            while name + '_' + str(n) in __names__: n += 1
            name = name + '_' + str(n) 
            self._name = name
        __names__.add(self._name)
        return self

    def __hdl_name__(self):
        if len(self._name) == 0: 
            self._setname()
        return self._name  

    def __hdl_emit__(self):
        return str(self.value)

class Node(HDL):
    ''' 
    hardware that outputs relays on inputs
    no operators on Node is allowed!
    return signal only
    '''
    def __init__(self):
        super().__init__()
        self.__hdl_visit__:int = 0       # travel DAG 
        self.inputs: List[Signal] = []   # wires
        self.outputs: List[Signal] = []  # wires
    
    def __hdl_name__(self): return self.__class__.__name__  
    def __hdl_emit__(self):
        left = ','.join([i.__hdl_name__() for i in self.outputs])
        right = ','.join([i.__hdl_name__() for i in self.inputs])
        return left + ' = ' + self.__hdl_name__() + '(' + right + ')' 

    def __new__(cls, *args, **kwargs):
        # return a signal, not node
        node = object.__new__(cls)
        node.__init__(*args, **kwargs)
        return node.outputs[0]


    def forward(self):  raise Exception("not implemented")  
    def step(self):     raise Exception("not implemented")  
        




from pyhgl.hdl._hdl import *
from pyhgl.hdl.hardware import * 
from pyhgl.hdl.signal import * 
from pyhgl.hdl._condition import *
from pyhgl.hdl.container import *
from pyhgl.hdl import _global





''' 
some decorators
'''

def bundle(f):
    def wrapper(*args, **kwargs):
        drop_list = f.__code__.co_varnames[:f.__code__.co_argcount]
        local_list = f(*args, **kwargs)
        new_dict = {}
        for k, v in local_list.items():
            if (k[0] != '_') and (k not in drop_list) and isinstance(v, (Signal, Container)):
                new_dict[k] = v 
                
        return Bundle(**new_dict) 
    return wrapper
        





def enum(cls):
    ret = {}
    for k, v in cls.__dict__.items():
        if k[0] != '_':
            ret[k] = v 
    n = len(ret)
    width = utils.width_infer(n-1)
    list_key = list(ret.keys())
    for i in range(n):
        ret[list_key[i]] = UInt(i, w = width)
    return Bundle(**ret)






''' 
inout wire only allowed to connect to tri-gate or inout wire 

input wire cannot appear at the left side

'''


def Input(signal: Signal):
    def toInput(atom_signal: Signal):
        ret = atom_signal._new()
        ret._direction = 'input'  
        return ret 
    return Map(signal, f = toInput)

def Output(signal: Signal):
    def toOutput(atom_signal: Signal):
        ret = atom_signal._new()
        ret._direction = 'output' 
        return ret 
    return Map(signal, f = toOutput)

def InOut(signal: Signal):
    def toInOut(atom_signal: Signal):
        ret = atom_signal._new()
        ret._direction = 'inout' 
        return ret 
    return Map(signal, f = toInOut)

def Flip(signal: Signal):
    '''
    do not return copy since _direction may setted 
    '''
    def setflip(atom_signal: Signal):
        atom_signal._flip = not ret._flip
        return atom_signal 
    return Map(signal, f = setflip)


def IO(io: Bundle) -> Bundle:
    ''' 
    apply flip 
    gen wire nodes
    return bundle
    '''
    def f(x: Signal):
        assert x._direction != 'inner', 'io signals shound have direction!'
        ret = Wire(x._new())
        ret._direction = x._direction
        if x._flip:
            if ret._direction == 'input':
                ret._direction = 'output' 
            if ret._direction == 'output':
                ret._direction = 'input'
        return ret
    return Map(io, f = f)



class _Clocks:
    ''' 
    add a clock: Clock.clock50MHz = 5e7
    '''
    def __init__(self):
        self._frequency: Dict[Signal, float] = {}
    def __setattr__(self, name, value):
        if name[0] == '_':
            object.__setattr__(self, name, value)
        elif hasattr(self, name):
            raise Exception("{name} already exists")
        else:
            frequency = float(value)
            object.__setattr__(self, name, Signal()._setname(name))
            self._frequency[getattr(self, name)] = frequency
    def _step(self):
        for clock in self._frequency.keys():
            clock.value = ~clock.value
    def __str__(self):
        ret = ''
        for signal, freq in self._frequency.items():
            ret += f"{signal._name}: freq = {freq} value = {signal.value}\n"
        return ret
        
_global._clocks = _Clocks()
_global._clocks.clock = 5e7
_global._reset = Signal()._setname('reset')

    


class Module(HDL):

    def __init__(self):
        ...
        
    def __post__(self):
        caller = sys._getframe(1).f_locals
        assert 'io' in caller and isinstance(caller['io'], Bundle), "module without io"
        for k, v in caller.items():
            if k[0] != '_' and isinstance(v, (Container, Signal)):
                v._setname(k)
                self.__dict__[k] = v
        # reverse direction
        for i in Flatten(self.io):
            if i._direction == 'input':
                i._direction = 'output'
            elif i._direction == 'output':
                i._direction = 'input'
        
class Circuit(HDL):
    def __init__(self, module):
        
        self.outputs: List[Signal] = []   # only consider output signal
        self.regs: List[Reg] = []
        self.comb_groups: List[List[Node]] = []
        
        for i in Flatten(module.io):
            if i._direction in ['input', 'inout']:
                self.outputs.append(i)
        
        self.rest_signal: List[Signal] = self.outputs[:]
        while len(self.rest_signal) > 0:
            temp = self.rest_signal
            self.rest_signal = []
            self.comb_groups.append([])
            self.curr_comb = self.comb_groups[-1]
            for node in temp:
                self.dfs(node)
            
    def dfs(self, node: Union[Signal, Node]):
        if isinstance(node, Signal):
            if node.driver is None:
                return 
            self.dfs(node.driver)
        elif abs(node.__hdl_visit__) != id(self):   # new  
            if isinstance(node, Reg):
                self.regs.append(node)
                self.rest_signal += node.inputs
                node.__hdl_visit__ = -id(self)      # regard as leaf node, visited
                return 
            else:  # is comb
                node.__hdl_visit__ = id(self)       # visiting
                for subnode in node.inputs:
                    self.dfs(subnode)
                node.__hdl_visit__ = -id(self)      # visited
                self.curr_comb.append(node)
                return 
        else:                                       
            if node.__hdl_visit__ == id(self):      # visiting
                raise Exception("loop detected in combinational logic")
            else:                                   # visited
                return


    def step(self):
        for i in self.regs:
            i.step()
        _global._clocks._step()
        for i in reversed(self.comb_groups):
            for j in i:
                j.forward()
        for i in self.regs:
            i.forward()


    def show(self):
        for i in self.regs:
            print(i, end='')
        for i in reversed(self.comb_groups):
            for j in i:
                print(j, end='')
        for i in self.regs:
            print(i.__hdl_emit2__())


""" 
TODO emit Module level IR (cannot detect circles)
"""




from .Logic import *



class UInt(Logic):
    ''' 
    '''
    def __init__(self,value: Union[int, str, bool] = 0, w: int = None):
        super().__init__()
        # basic attribute
        self.value: int = 0 
        self.width: int = 1 
        self.shape: Tuple[int, ...] = None
        # leaf node ? 
        self.driver = None   
        self._name = ""
        # for input output 
        self._flip = False 
        self._direction = 'inner'
        
        if isinstance(value, (bool,int)):
            self.value = int(value)
            self.width = utils.width_infer(self.value)
        elif isinstance(value, str):
            assert value[0] not in ['$', '+', '-'], 'UInt not accept signed value'
            self.value, self.width, _ = utils.str2int(value)
        else:
            raise Exception("invalid signal input")

        if w is not None:
            assert w >= self.width, f'overflow: UInt({value}) with width {w}'
            self.width = w
    
    


"""modified cpython tokenize.py
"""

import collections
import functools
import itertools as _itertools
import re
import sys
import io

import pyhgl.config

token_def = """\
ENDMARKER
NAME
NUMBER
STRING
NEWLINE
INDENT
DEDENT

LPAR                    '('
RPAR                    ')'
LSQB                    '['
RSQB                    ']'
COLON                   ':'
COMMA                   ','
SEMI                    ';'
PLUS                    '+'
MINUS                   '-'
STAR                    '*'
SLASH                   '/'
VBAR                    '|'
AMPER                   '&'
LESS                    '<'
GREATER                 '>'
EQUAL                   '='
DOT                     '.'
PERCENT                 '%'
LBRACE                  '{'
RBRACE                  '}'
EQEQUAL                 '=='
NOTEQUAL                '!='
LESSEQUAL               '<='
GREATEREQUAL            '>='
TILDE                   '~'
CIRCUMFLEX              '^'
LEFTSHIFT               '<<'
RIGHTSHIFT              '>>'
DOUBLESTAR              '**'
PLUSEQUAL               '+='
MINEQUAL                '-='
STAREQUAL               '*='
SLASHEQUAL              '/='
PERCENTEQUAL            '%='
AMPEREQUAL              '&='
VBAREQUAL               '|='
CIRCUMFLEXEQUAL         '^='
LEFTSHIFTEQUAL          '<<='
RIGHTSHIFTEQUAL         '>>='
DOUBLESTAREQUAL         '**='
DOUBLESLASH             '//'
DOUBLESLASHEQUAL        '//='
AT                      '@'
ATEQUAL                 '@='
RARROW                  '->'
ELLIPSIS                '...'
COLONEQUAL              ':='

# pyhgl operators
LELELE                  '<<<'
GEGEGE                  '>>>'
LEEQEQ                  '<=='
EQEQGE                  '==>'
LEEQGE                  '<=>'

OP
AWAIT
ASYNC
TYPE_IGNORE
TYPE_COMMENT
SOFT_KEYWORD
ERRORTOKEN

# These aren't used by the C tokenizer but are needed for tokenize.py
COMMENT
NL
ENCODING
"""

token_list = []   # (name, value, string)
for line in token_def.splitlines():
    # strip comments
    i = line.find('#')
    if i >= 0:
        line = line[:i]
    line = line.strip()
    if not line:
        continue 
    line = line.split()
    name = line[0]
    value = len(token_list)
    string = eval(line[1]) if len(line) > 1 else None   # "'='" -> '=' 
    token_list.append((name, value, string))

token_list.append(("N_TOKENS", len(token_list), None))  # number of tokens 
token_list.append(("NT_OFFSET", 256, None))             # Special definitions for cooperation with parser

tok_name = {}                   # {0:'ENDMARKER', 1:'NAME', ...}
EXACT_TOKEN_TYPES = {}          # {'(': LPAR, ')': RPAR, ...}

for name, value, string in token_list:
    globals()[name] = value     #  ENDMARKER = 0, NAME = 1, ...
    tok_name[value] = name 
    if string:
        EXACT_TOKEN_TYPES[string] = value


def ISTERMINAL(x):
    return x < NT_OFFSET

def ISNONTERMINAL(x):
    return x >= NT_OFFSET

def ISEOF(x):
    return x == ENDMARKER


cookie_re = re.compile(r'^[ \t\f]*#.*?coding[:=][ \t]*([-\w.]+)', re.ASCII)
blank_re = re.compile(br'^[ \t\f]*(?:[#\r\n]|$)', re.ASCII)



class TokenInfo(collections.namedtuple('TokenInfo', 'type string start end line')): # type is a number
    def __repr__(self):
        annotated_type = "{:<3}: {:<10}".format(self.type, tok_name[self.type])  
        return "TokenInfo: start={!r:<10} end={!r:<10} {} string: {!r:<10} line: {!r}".format(
            self.start, self.end, annotated_type, self.string,  self.line) 

            

def group(*choices): return '(' + '|'.join(choices) + ')'
def any(*choices): return group(*choices) + '*'
def maybe(*choices): return group(*choices) + '?'

# Note: we use unicode matching for names ("\w") but ascii matching for
# number literals.
Whitespace = r'[ \f\t]*'
Comment = r'#[^\r\n]*'
Ignore = Whitespace + any(r'\\\r?\n' + Whitespace) + maybe(Comment)
Name = r'\w+'

Hexnumber = r'0[xX](?:_?[0-9a-fA-F])+'
Binnumber = r'0[bB](?:_?[01])+'
Octnumber = r'0[oO](?:_?[0-7])+'
Decnumber = r'(?:0(?:_?0)*|[1-9](?:_?[0-9])*)'
Intnumber = group(Hexnumber, Binnumber, Octnumber, Decnumber)
Exponent = r'[eE][-+]?[0-9](?:_?[0-9])*'
Pointfloat = group(r'[0-9](?:_?[0-9])*\.(?:[0-9](?:_?[0-9])*)?',
                   r'\.[0-9](?:_?[0-9])*') + maybe(Exponent)
Expfloat = r'[0-9](?:_?[0-9])*' + Exponent
Floatnumber = group(Pointfloat, Expfloat)
Imagnumber = group(r'[0-9](?:_?[0-9])*[jJ]', Floatnumber + r'[jJ]')
Number = group(Imagnumber, Floatnumber, Intnumber)

# Return the empty string, plus all of the valid string prefixes.
def _all_string_prefixes():
    # The valid string prefixes. Only contain the lower case versions,
    #  and don't contain any permutations (include 'fr', but not
    #  'rf'). The various permutations will be generated.
    _valid_string_prefixes = ['b', 'r', 'u', 'f', 'br', 'fr']
    # if we add binary f-strings, add: ['fb', 'fbr']
    result = {''}
    for prefix in _valid_string_prefixes:
        for t in _itertools.permutations(prefix):
            # create a list with upper and lower versions of each
            #  character
            for u in _itertools.product(*[(c, c.upper()) for c in t]):
                result.add(''.join(u))
    return result

@functools.lru_cache
def _compile(expr):
    return re.compile(expr, re.UNICODE)

# Note that since _all_string_prefixes includes the empty string,
#  StringPrefix can be the empty string (making it optional).
StringPrefix = group(*_all_string_prefixes())

# Tail end of ' string.
Single = r"[^'\\]*(?:\\.[^'\\]*)*'"
# Tail end of " string.
Double = r'[^"\\]*(?:\\.[^"\\]*)*"'
# Tail end of ''' string.
Single3 = r"[^'\\]*(?:(?:\\.|'(?!''))[^'\\]*)*'''"
# Tail end of """ string.
Double3 = r'[^"\\]*(?:(?:\\.|"(?!""))[^"\\]*)*"""'
Triple = group(StringPrefix + "'''", StringPrefix + '"""')
# Single-line ' or " string.
String = group(StringPrefix + r"'[^\n'\\]*(?:\\.[^\n'\\]*)*'",
               StringPrefix + r'"[^\n"\\]*(?:\\.[^\n"\\]*)*"')

# Sorting in reverse order puts the long operators before their prefixes.
# Otherwise if = came before ==, == would get recognized as two instances
# of =.
Special = group(*map(re.escape, sorted(EXACT_TOKEN_TYPES, reverse=True)))
Funny = group(r'\r?\n', Special)

PlainToken = group(Number, Funny, String, Name)
Token = Ignore + PlainToken

# First (or only) line of ' or " string.
ContStr = group(StringPrefix + r"'[^\n'\\]*(?:\\.[^\n'\\]*)*" +
                group("'", r'\\\r?\n'),
                StringPrefix + r'"[^\n"\\]*(?:\\.[^\n"\\]*)*' +
                group('"', r'\\\r?\n'))
PseudoExtras = group(r'\\\r?\n|\Z', Comment, Triple)
PseudoToken = Whitespace + group(PseudoExtras, Number, Funny, ContStr, Name)

# For a given string prefix plus quotes, endpats maps it to a regex
#  to match the remainder of that string. _prefix can be empty, for
#  a normal single or triple quoted string (with no prefix).
endpats = {}
for _prefix in _all_string_prefixes():
    endpats[_prefix + "'"] = Single
    endpats[_prefix + '"'] = Double
    endpats[_prefix + "'''"] = Single3
    endpats[_prefix + '"""'] = Double3

# A set of all of the single and triple quoted string prefixes,
#  including the opening quotes.
single_quoted = set()
triple_quoted = set()
for t in _all_string_prefixes():
    for u in (t + '"', t + "'"):
        single_quoted.add(u)
    for u in (t + '"""', t + "'''"):
        triple_quoted.add(u)

tabsize = 8

class TokenError(Exception): pass

class StopTokenizing(Exception): pass




def _tokenize(readline):
    
    if type(readline) is str:
        readline = io.StringIO(readline).readline
    
    lnum = parenlev = continued = 0
    numchars = '0123456789'
    contstr, needcont = '', 0
    contline = None
    indents = [0]

    last_line = b''
    line = b''
    while True:                                # loop over lines in stream
        last_line = line
        line = readline()

        lnum += 1
        pos, max = 0, len(line)

        if contstr:                            # continued string
            if not line:
                raise TokenError("EOF in multi-line string", strstart)
            endmatch = endprog.match(line)
            if endmatch:
                pos = end = endmatch.end(0)
                yield TokenInfo(STRING, contstr + line[:end],
                       strstart, (lnum, end), contline + line)
                contstr, needcont = '', 0
                contline = None
            elif needcont and line[-2:] != '\\\n' and line[-3:] != '\\\r\n':
                yield TokenInfo(ERRORTOKEN, contstr + line,
                           strstart, (lnum, len(line)), contline)
                contstr = ''
                contline = None
                continue
            else:
                contstr = contstr + line
                contline = contline + line
                continue

        elif parenlev == 0 and not continued:  # new statement
            if not line: break
            column = 0
            while pos < max:                   # measure leading whitespace
                if line[pos] == ' ':
                    column += 1
                elif line[pos] == '\t':
                    column = (column//tabsize + 1)*tabsize
                elif line[pos] == '\f':
                    column = 0
                else:
                    break
                pos += 1
            if pos == max:
                break

            if line[pos] in '#\r\n':           # skip comments or blank lines
                if line[pos] == '#':
                    comment_token = line[pos:].rstrip('\r\n')
                    yield TokenInfo(COMMENT, comment_token,
                           (lnum, pos), (lnum, pos + len(comment_token)), line)
                    pos += len(comment_token)

                yield TokenInfo(NL, line[pos:],
                           (lnum, pos), (lnum, len(line)), line)
                continue

            if column > indents[-1]:           # count indents or dedents
                indents.append(column)
                yield TokenInfo(INDENT, line[:pos], (lnum, 0), (lnum, pos), line)
            while column < indents[-1]:
                if column not in indents:
                    raise IndentationError(
                        "unindent does not match any outer indentation level",
                        ("<tokenize>", lnum, pos, line))
                indents = indents[:-1]

                yield TokenInfo(DEDENT, '', (lnum, pos), (lnum, pos), line)

        else:                                  # continued statement
            if not line:
                raise TokenError("EOF in multi-line statement", (lnum, 0))
            continued = 0

        while pos < max:
            pseudomatch = _compile(PseudoToken).match(line, pos)
            if pseudomatch:                                # scan for tokens
                start, end = pseudomatch.span(1)
                spos, epos, pos = (lnum, start), (lnum, end), end
                if start == end:
                    continue
                token, initial = line[start:end], line[start]

                if (initial in numchars or                 # ordinary number
                    (initial == '.' and token != '.' and token != '...')):
                    yield TokenInfo(NUMBER, token, spos, epos, line)
                elif initial in '\r\n':
                    if parenlev > 0:
                        yield TokenInfo(NL, token, spos, epos, line)
                    else:
                        yield TokenInfo(NEWLINE, token, spos, epos, line)

                elif initial == '#':
                    assert not token.endswith("\n")
                    yield TokenInfo(COMMENT, token, spos, epos, line)

                elif token in triple_quoted:
                    endprog = _compile(endpats[token])
                    endmatch = endprog.match(line, pos)
                    if endmatch:                           # all on one line
                        pos = endmatch.end(0)
                        token = line[start:pos]
                        yield TokenInfo(STRING, token, spos, (lnum, pos), line)
                    else:
                        strstart = (lnum, start)           # multiple lines
                        contstr = line[start:]
                        contline = line
                        break

                # Check up to the first 3 chars of the token to see if
                #  they're in the single_quoted set. If so, they start
                #  a string.
                # We're using the first 3, because we're looking for
                #  "rb'" (for example) at the start of the token. If
                #  we switch to longer prefixes, this needs to be
                #  adjusted.
                # Note that initial == token[:1].
                # Also note that single quote checking must come after
                #  triple quote checking (above).
                elif (initial in single_quoted or
                      token[:2] in single_quoted or
                      token[:3] in single_quoted):
                    if token[-1] == '\n':                  # continued string
                        strstart = (lnum, start)
                        # Again, using the first 3 chars of the
                        #  token. This is looking for the matching end
                        #  regex for the correct type of quote
                        #  character. So it's really looking for
                        #  endpats["'"] or endpats['"'], by trying to
                        #  skip string prefix characters, if any.
                        endprog = _compile(endpats.get(initial) or
                                           endpats.get(token[1]) or
                                           endpats.get(token[2]))
                        contstr, needcont = line[start:], 1
                        contline = line
                        break
                    else:                                  # ordinary string
                        yield TokenInfo(STRING, token, spos, epos, line)

                elif initial.isidentifier():               # ordinary name
                    yield TokenInfo(NAME, token, spos, epos, line)
                elif initial == '\\':                      # continued stmt
                    continued = 1
                else:
                    if initial in '([{':
                        parenlev += 1
                    elif initial in ')]}':
                        parenlev -= 1
                    yield TokenInfo(OP, token, spos, epos, line)
            else:
                yield TokenInfo(ERRORTOKEN, line[pos],
                           (lnum, pos), (lnum, pos+1), line)
                pos += 1

    # Add an implicit NEWLINE if the input doesn't end in one
    if last_line and last_line[-1] not in '\r\n' and not last_line.strip().startswith("#"):
        yield TokenInfo(NEWLINE, '', (lnum - 1, len(last_line)), (lnum - 1, len(last_line) + 1), '')
    for indent in indents[1:]:                 # pop remaining indent levels
        yield TokenInfo(DEDENT, '', (lnum, 0), (lnum, 0), '')
    yield TokenInfo(ENDMARKER, '', (lnum, 0), (lnum, 0), '')





import logging 

logger = logging.getLogger('pyhgl')
ch = logging.StreamHandler()
ch.setFormatter(logging.Formatter("[%(pathname)s:%(lineno)d] %(message)s "))
logger.addHandler(ch)



hook_verbose = False
tokenize_verbose = False
parser_verbose = False
